# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: tasks.proto
# plugin: python-betterproto
# This file has been @generated

from dataclasses import dataclass
from typing import (
    List,
    Optional,
)

import betterproto

from protobunny import models as models

from ... import commons as __commons__


@dataclass(eq=False, repr=False)
class TaskMessage(models.ProtoBunnyMessage):
    content: str = betterproto.string_field(10)
    weights: List[float] = betterproto.float_field(30)
    bbox: List[int] = betterproto.int64_field(40)
    options: Optional["__commons__.JsonContent | dict"] = betterproto.message_field(
        50, optional=True
    )
