/* This .proto file contains messages definitions shared among the packages */
syntax = "proto3";

package commons;

/* The JsonContent message is used as a field in other messages to hold dictionary/Json data.
Notes:
Serialization/deserialization of pb.commons.JsonContent fields happens transparently
(no need to build the JsonContent message, just set the field with the dictionary).

Supported betterproto methods: __bytes__, parse, to_dict, from_dict, to_pydict, to_json
*/
message JsonContent {
  bytes content = 1;
}
