# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: tests.proto
# plugin: python-betterproto
# This file has been @generated

from dataclasses import dataclass
from typing import Optional

import betterproto

from protobunny import models as models

from .. import commons as _commons__


class Color(betterproto.Enum):
    OFF = 0
    GREEN = 1
    ORANGE = 2
    GREEN_ORANGE = 3
    RED = 4
    RED_GREEN = 5
    RED_ORANGE = 6
    RED_ORANGE_GREEN = 7


@dataclass(eq=False, repr=False)
class TestMessage(models.ProtoBunnyMessage):
    content: str = betterproto.string_field(10)
    number: int = betterproto.int64_field(20)
    detail: Optional[str] = betterproto.string_field(30, optional=True)
    options: Optional["_commons__.JsonContent | dict"] = betterproto.message_field(
        40, optional=True
    )
    color: Optional["Color"] = betterproto.enum_field(50, optional=True)


@dataclass(eq=False, repr=False)
class ComplexTestMessage(models.ProtoBunnyMessage):
    x: int = betterproto.int64_field(10)
    y: Optional[int] = betterproto.int64_field(20, optional=True)
    a: bool = betterproto.bool_field(1, group="group")
    b: int = betterproto.int32_field(2, group="group")
    c: str = betterproto.string_field(3, group="group")


#######################################################
# Dynamically added by protobunny post_compile.py

from . import tasks  # noqa
