# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: protobunny/commons.proto
# plugin: python-betterproto
# This file has been @generated

from dataclasses import dataclass

import betterproto

from protobunny import models as models


@dataclass(eq=False, repr=False)
class JsonContent(models.ProtoBunnyMessage):
    """
    The JsonContent message is used as a field in other messages to hold dictionary/Json data.
    Notes:
    Serialization/deserialization of pb.commons.JsonContent fields happens transparently
    (no need to build the JsonContent message, just set the field with the dictionary).

    Supported betterproto methods: __bytes__, parse, to_dict, from_dict, to_pydict, to_json
    """

    content: bytes = betterproto.bytes_field(1)
